#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 2/21/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Iterable, Dict, Any

# External modules
import numpy as np
import xarray as xr

# Internal modules


logger = logging.getLogger(__name__)


def create_vgrid(
        ds: xr.Dataset,
        vcoords: Iterable[str]
) -> xr.Dataset:
    """
    Create a vertical grid based on given vertical coordinates
    """
    ds = ds.copy()
    avail_vcoords = [c for c in vcoords if c in ds.dims]
    vgrid = np.concatenate([ds[c].values for c in avail_vcoords])
    ds['vgrid'] = vgrid
    return ds


def add_no_vgrid(
        ds: xr.Dataset,
        vcoords: Iterable[str],
        val: float = 0
) -> xr.Dataset:
    """
    Add an additional variable if no vertical grid is available
    """
    ds = ds.copy()
    vars_wo_vgrid = [var for var in ds.data_vars
                     if set(ds[var].dims).isdisjoint(vcoords)]
    for var in vars_wo_vgrid:
        ds[var] = ds[var].expand_dims('no_vgrid', axis=-3)
    if vars_wo_vgrid:
        ds['no_vgrid'] = np.array([val, ])
    return ds


def replace_grid(
        ds: xr.Dataset,
        remap_dict: Dict[str, Any]
) -> xr.Dataset:
    """
    Replace grid variables within given dataset
    """
    rename_dict = {k: v for k, v in remap_dict.items()
                   if k in ds.coords}
    ds = ds.drop_sel(list(rename_dict.keys()))
    ds = ds.rename(rename_dict)
    return ds


def ds_to_array(ds: xr.Dataset, grid_dims: Iterable[str]) -> xr.DataArray:
    """
    Stack a prepared dataset into a valid state array
    """
    unified_data = ds.to_array(dim='var_name')
    stacked_data = unified_data.stack(grid=grid_dims)
    if 'ensemble' not in stacked_data.dims:
        stacked_data = stacked_data.expand_dims(dim='ensemble')
    transpose_dims = ['var_name', 'time', 'ensemble', 'grid']
    transpose_dims = [d for d in transpose_dims if d in stacked_data.dims]
    prepared_data = stacked_data.transpose(*transpose_dims)
    return prepared_data


def array_to_ds(data: xr.DataArray) -> xr.Dataset:
    """
    Unstack an array to dataset
    """
    unstacked_data = data.unstack('grid')
    grid_dims = list(data.indexes['grid'].names)
    transpose_dims = ['var_name', 'ensemble', 'time'] + list(grid_dims)
    transpose_dims = [d for d in transpose_dims if d in unstacked_data.dims]
    transposed_data = unstacked_data.transpose(*transpose_dims)
    prepared_ds = transposed_data.to_dataset(dim='var_name')
    return prepared_ds


def dim_transpose(array: xr.DataArray, vcoords: Iterable[str]):
    dim_generic = ['time', 'ensemble']
    dim_order = [d for d in dim_generic if d in array.dims]
    dim_order += [d for d in vcoords+['vgrid', ] if d in array.dims]
    dim_grid = [d for d in array.dims if d not in dim_order]
    dim_order += dim_grid
    array_trans = array.transpose(*dim_order)
    return array_trans


def generic_postprocess(
        analysis_data: xr.DataArray,
        origin_ds: xr.Dataset,
        vcoords: Iterable[str]
) -> xr.Dataset:
    """
    This function can be used to post-process analysis data and incorporate
    included variables into given origin dataset. There are different steps
    included:

    1. Unstack the grid form analysis data and convert the data into a dataset
    2. Iterate through included variables and reindex these variables as they
        are in given origin dataset
    3. Replace the variables in given origin dataset with the analysis variables

    Parameters
    ----------
    analysis_data : :py:class:`xarray.DataArray`
        This array represents a valid analysis state, which can be a result of
        any given assimilation algorithm included in this package. In this
        analysis data, the horizontal and vertical grid are stacked together and
        the variables are a dimension within the array.
    origin_ds : :py:class:`xarray.Dataset`
        This origin dataset is used as source dataset to convert given analysis
        array into a valid origin dataset. The resulting dataset is a copy of
        this dataset, where the assimilated variables are replaced.
    vcoords : Iterable(str)
        These vcoords are used to unstack the analysis data.

    Returns
    -------
    analysis_ds : :py:class:`xarray.Dataset`
        This analysis dataset is a copy of given origin dataset with replaced
        variables from given analysis array.
    """
    pre_analysis_ds = array_to_ds(analysis_data)
    logger.info('Created analysis dataset')
    analysis_ds = origin_ds.copy(deep=True)
    logger.info('Copied original dataset')
    for var in pre_analysis_ds.data_vars:
        logger.info('Starting to post-process {0:s}'.format(var))
        try:
            data_prepared = pre_analysis_ds[var].dropna('vgrid', how='all')
            data_prepared = dim_transpose(data_prepared, vcoords)
            tmp_analysis_var = dim_transpose(analysis_ds[var], vcoords)
            tmp_analysis_var.data = data_prepared.data.reshape(
                tmp_analysis_var.shape
            )
            analysis_ds[var] = tmp_analysis_var.transpose(
                *analysis_ds[var].dims
            )
            logger.info('Post-processed {0:s}'.format(var))
        except KeyError:
            logger.warning('Var: {0:s} is not found'.format(var))
        except ValueError:
            logger.warning(
                'Var: {0:s} is not broadcastable ({1:s} != {2:s})'.format(
                    var, str(data_prepared.shape),
                    str(analysis_ds[var].shape)
                )
            )
    return analysis_ds



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch

# Internal modules
from .base_kernels import BaseKernel
from . import utils as kernel_utils


logger = logging.getLogger(__name__)


__all__ = ['LinearKernel']


class LinearKernel(BaseKernel):
    """
    This linear kernel specifies a dot product between given matrices. The
    linear kernel can be used to construct a Bayesian linear regression. The
    kernelized ensemble transform Kalman filter (
    :py:class:`pytassim.assimilation.filter.ketkf.KETKFUncorr`) with this
    kernel reconstructs the standard ensemble transform Kalman filter (
    :py:class:`pytassim.assimilation.filter.etkf.ETKFUncorr`). This kernel is
    not stationary and degenerated, meaning that it is not translation-invariant
    and might lead to inconsistent results in non-linear regimes.


    .. math::

       K(x_i, x_j) = {(x_i)}^{T} x_j

    """
    def __str__(self) -> str:
        return 'LinearKernel'

    def __repr__(self) -> str:
        return 'Linear'

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return kernel_utils.dot_product(x, y)



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/6/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc
from typing import Union, Iterable, Tuple, List

# External modules
import xarray as xr
import numpy as np
import pandas as pd
import torch

# Internal modules
from .etkf_core import ETKFAnalyser, CorrMixin, UnCorrMixin
from .filter import FilterAssimilation
from pytassim.transform import BaseTransformer


logger = logging.getLogger(__name__)


__all__ = [
    'ETKFCorr',
    'ETKFUncorr'
]


class ETKFBase(FilterAssimilation):
    """
    The base object for the ensemble transform Kalman filter.
    """
    def __init__(
            self,
            inf_factor: Union[float, torch.Tensor, torch.nn.Parameter] = 1.0,
            smoother: bool = False, gpu: bool = False,
            pre_transform: Union[None, Iterable[BaseTransformer]] = None,
            post_transform: Union[None, Iterable[BaseTransformer]] = None
    ):
        super().__init__(smoother=smoother, gpu=gpu,
                         pre_transform=pre_transform,
                         post_transform=post_transform)
        self._name = 'Global ETKF'
        self._weights = None
        self.inf_factor = inf_factor

    def __str__(self):
        return 'Global ETKF({0})'.format(self.inf_factor)

    def __repr__(self):
        return 'ETKF'

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self._analyser.inf_factor

    @inf_factor.setter
    def inf_factor(
            self,
            new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        """
        Sets a new inflation factor.
        """
        self._analyser = ETKFAnalyser(inf_factor=new_factor)

    @property
    def analyser(self) -> ETKFAnalyser:
        return self._analyser

    @property
    def weights(self) -> torch.Tensor:
        return self._weights

    def _normalise_obs(
            self,
            pseudo_obs: torch.Tensor,
            obs: torch.Tensor,
            cinv: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Normalise given pseudo observations and observations with given
        inverse cholesky decomposition of the observational covariance.
        """
        pseudo_mean = pseudo_obs.mean(dim=-2, keepdim=True)
        normed_perts = self._mul_cinv(pseudo_obs-pseudo_mean, cinv)
        normed_obs = self._mul_cinv(obs.view(1, -1)-pseudo_mean, cinv)
        return normed_perts, normed_obs

    @abc.abstractmethod
    def _mul_cinv(
            self,
            state: torch.Tensor,
            cinv: torch.Tensor
    ) -> torch.Tensor:
        """
        Multiply given state with given inverse cholesky decomposition of the
        observational covariance.
        """
        pass

    @abc.abstractmethod
    def _get_chol_inverse(self, cov: torch.Tensor) -> torch.Tensor:
        """
        Estimate the cholesky inverse of given covariance matrix.
        """
        pass

    def update_state(
            self,
            state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]],
            pseudo_state: xr.DataArray,
            analysis_time: pd.Timestamp
    ) -> xr.DataArray:
        """
        This method updates the state based on given observations and analysis
        time. This method prepares the different states, calculates the ensemble
        weights and applies these weight to given state. The calculation of the
        weights is based on PyTorch, while everything else is calculated with
        Numpy / Xarray.

        Parameters
        ----------
        state : :py:class:`xarray.DataArray`
            This state is updated by this assimilation algorithm and given
            ``observation``. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`.
        observations : :py:class:`xarray.Dataset` or \
        iterable(:py:class:`xarray.Dataset`)
            These observations are used to update given state. An iterable of
            many :py:class:`xarray.Dataset` can be used to assimilate different
            variables. For the observation state, these observations are
            stacked such that the observation state contains all observations.
        pseudo_state : :py:class:`xarray.DataArray`
            This state is used to generate an observation-equivalent. This
             :py:class:`~xarray.DataArray` should have four coordinates, which
             are specified in :py:class:`pytassim.state.ModelState`.
        analysis_time : :py:class:`datetime.datetime`
            This analysis time determines at which point the state is updated.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            The analysed state based on given state and observations. The
            analysis has same coordinates as given ``state``. If filtering mode
            is on, then the time axis has only one element.
        """
        logger.info('####### {0:s} #######'.format(self._name))
        logger.info('Starting with specific preparation')
        pseudo_obs, obs_state, obs_cov, obs_grid = self._get_states(
            pseudo_state, observations,
        )

        logger.info('Transfering the data to torch')
        pseudo_obs, obs_state, obs_cov = self._states_to_torch(
            pseudo_obs, obs_state, obs_cov
        )

        logger.info('Normalise perturbations and observations')
        obs_cinv = self._get_chol_inverse(obs_cov)
        normed_perts, normed_obs = self._normalise_obs(pseudo_obs, obs_state,
                                                       obs_cinv)
        state_mean, state_perts = state.state.split_mean_perts()
        state_grid = state_perts.indexes['grid']
        state_perts, = self._states_to_torch(state_perts.values)

        logger.info('Create analysis perturbations')
        analysis_perts = self.analyser(state_perts, normed_perts, normed_obs,
                                       state_grid, obs_grid)

        logger.info('Create analysis')
        analysis_perts = state.copy(data=analysis_perts.numpy())
        analysis = analysis_perts + state_mean
        analysis = analysis.transpose('var_name', 'time', 'ensemble', 'grid')
        return analysis

    def _get_states(
            self,
            pseudo_state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]]
    ) -> Union[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        This method prepares the different parts of the state. It calculates
        statistics in observation space and concatenates given observations into
        a long vector. Observations without an observation operator, defined in
        :py:meth:`xarray.Dataset.obs.operator`, are skipped. This method
        prepares the states in Numpy / Xarray.

        Parameters
        ----------
        pseudo_state : :py:class:`xarray.DataArray`
            This state is used to generate an observation-equivalent. It is
            further updated by this assimilation algorithm and given
            ``observation``. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`.
        observations : :py:class:`xarray.Dataset` or \
        iterable(:py:class:`xarray.Dataset`)
            These observations are used to update given state. An iterable of
            many :py:class:`xarray.Dataset` can be used to assimilate different
            variables. For the observation state, these observations are
            stacked such that the observation state contains all observations.

        Returns
        -------
        innov : :py:class:`numpy.ndarray`
            This vector contains the innovations, calculated with
            :math:`\\textbf{y}^{o} - \\overline{h(\\textbf{x}^{b})}`. The length
            of this vector is :math:`l`, the observation length.
        hx_perts : :py:class:`numpy.ndarray`
            This matrix contains the ensemble perturbations in ensemble space.
            The `i`-th perturbation is calculated based on
            :math:`h(\\textbf{x}_{i}^{b})-\\overline{h(\\textbf{x}^{b})}`. The
            shape of this matrix is :math:`l~x~k`, with :math:`k` as ensemble
            size and :math:`l` as observation length.
        obs_cov : :py:class:`numpy.ndarray`
            The concatenated observation covariance. This covariance is created
            with the assumption that different observation subset are not
            correlated. This matrix has a shape of :math:`l~x~l`, with :math:`l`
            as observation length.
        obs_grid : :py:class:`numpy.ndarray`
            This is the concatenated observation grid. This can be used for
            localization or weighting purpose. This last axis of this array has
            a length of :math:`l`, the observation length.
        """
        logger.info('Apply observation operator')
        pseudo_obs, filtered_obs = self._get_pseudo_obs(pseudo_state,
                                                        observations)
        logger.info('Concatenate observations')
        obs_state, obs_grid = self._prepare_obs(filtered_obs)
        obs_cov = self._get_obs_cov(filtered_obs)
        return pseudo_obs, obs_state, obs_cov, obs_grid

    def _get_pseudo_obs(
            self,
            state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]]
    ) -> Tuple[np.ndarray, List[np.ndarray]]:
        """
        Get pseudo observational array and filtered observations. This method
        applies the observation operator and concatenates the pseudo
        observations.
        """
        pseudo_obs, filtered_obs = self._apply_obs_operator(state, observations)
        pseudo_obs = self._cat_pseudo_obs(pseudo_obs)
        return pseudo_obs, filtered_obs

    @staticmethod
    def _cat_pseudo_obs(pseudo_obs: Iterable[xr.DataArray]) -> np.ndarray:
        """
        Concatenate given pseudo observations into a pseudo observational
        array.
        """
        state_stacked_list = []
        for obs in pseudo_obs:
            if isinstance(obs.indexes['obs_grid_1'], pd.MultiIndex):
                obs['obs_grid_1'] = pd.Index(
                    obs.indexes['obs_grid_1'].values, tupleize_cols=False
                )
            stacked_obs = obs.stack(obs_id=('time', 'obs_grid_1'))
            state_stacked_list.append(stacked_obs)
        pseudo_obs_concat = xr.concat(state_stacked_list, dim='obs_id')
        pseudo_obs_concat = pseudo_obs_concat.data
        return pseudo_obs_concat


class ETKFCorr(CorrMixin, ETKFBase):
    """
    This is an implementation of the `ensemble transform Kalman filter`
    :cite:`bishop_adaptive_2001` for correlated observation covariances.
    This ensemble Kalman filter is a deterministic filter, where the state is
    update globally. This ensemble Kalman filter estimates ensemble weights in
    weight space, which are then applied to the given state. This implementation
    follows :cite:`hunt_efficient_2007` with global weight estimation and is
    implemented in PyTorch.
    This implementation allows filtering in time based on linear propagation
    assumption :cite:`hunt_four-dimensional_2004` and ensemble smoothing.

    Parameters
    ----------
    smoother : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self) -> str:
        return 'Correlated {0:s}'.format(str(super(ETKFBase)))

    def __repr__(self) -> str:
        return 'Corr{0:s}'.format(repr(super(ETKFBase)))


class ETKFUncorr(UnCorrMixin, ETKFBase):
    """
    This is an implementation of the `ensemble transform Kalman filter`
    :cite:`bishop_adaptive_2001` for uncorrelated observation covariances.
    This ensemble Kalman filter is a deterministic filter, where the state is
    update globally. This ensemble Kalman filter estimates ensemble weights in
    weight space, which are then applied to the given state. This implementation
    follows :cite:`hunt_efficient_2007` with global weight estimation and is
    implemented in PyTorch.
    This implementation allows filtering in time based on linear propagation
    assumption :cite:`hunt_four-dimensional_2004` and ensemble smoothing.

    Parameters
    ----------
    smoother : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self) -> str:
        return 'Uncorrelated {0:s}'.format(str(super(ETKFBase)))

    def __repr__(self) -> str:
        return 'Uncorr{0:s}'.format(repr(super(ETKFBase)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# System modules
from setuptools import setup, find_packages
from codecs import open
from os import path
import glob

# External modules

# Internal modules


__version__ = "0.2.1"

BASE = path.abspath(path.dirname(__file__))

long_description = open(path.join(BASE, 'README.rst'), encoding='utf-8').read()


setup(
    name='torch-assimilate',

    version=__version__,

    description='torch-assimilate is a data assimilation package based on '
                'PyTorch, xarray and dask',
    long_description=long_description,
    long_description_content_type="text/x-rst",

    url='https://gitlab.com/tobifinn/torch-assimilate',

    author='Tobias Sebastian Finn',
    author_email='tobias.sebastian.finn@uni-hamburg.de',

    license='GPL3',

    keywords='statistics meteorology pre-processing testbed forecast '
             'assimilation data',

    packages=find_packages(exclude=['contrib', 'docs', 'tests.*', 'test']),

    test_suite='tests',
)



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 26.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from copy import deepcopy
from typing import Union, Iterable

# External modules
import torch.nn

import xarray as xr
import pandas as pd

# Internal modules
from ..base import BaseAssimilation
from pytassim.transform import BaseTransformer


logger = logging.getLogger(__name__)


class NeuralAssimilation(BaseAssimilation):
    """
    NeuralAssimilation is a class to assimilate observations into a state with
    neural networks. This class supports PyTorch natively. The ``assimilate``
    method of a given :py:class:`torch.nn.Module` is used to to assimilate the
    given state. No observation operator is needed for given observations.

    Parameters
    ----------
    model : child of :py:class:`torch.nn.Module`
        This model is used to assimilate given observations into given state.
        The model needs an ``assimilate`` method, where state, flattened
        observations and flattened observation covariance is given. This model
        is transferred to specified device. Attention, a new clone of this
        model is created.
    smoother : bool, optional
        This bool indicates if given `state` and `observations` should be
        localized to given `analysis_time`. If True, full state and all
        observations are passed to neural network. If False, localized state and
        nearest observations are only used. Default is True.
    gpu : bool, optional
        This boolean indicates if the assimilation should be done on GPU (True)
        or CPU (False). Default value is False, indicating computations on CPU.
        This flag transfers also given model to CPU or GPU.
    """
    def __init__(
            self,
            model: torch.nn.Module,
            smoother: bool = False, gpu: bool = False,
            pre_transform: Union[None, Iterable[BaseTransformer]] = None,
            post_transform: Union[None, Iterable[BaseTransformer]] = None
    ):
        super().__init__(smoother=smoother, gpu=gpu,
                         pre_transform=pre_transform,
                         post_transform=post_transform)
        self._model = None
        self.model = model

    def __str__(self) -> str:
        return 'NeuralAssimilation'

    def __repr__(self) -> str:
        return 'NeuralAssimilation'

    @property
    def model(self) -> torch.nn.Module:
        return self._model

    @model.setter
    def model(self, new_model: torch.nn.Module):
        if not hasattr(new_model, 'assimilate'):
            raise TypeError('Given model is not a valid assimilation model!')
        cloned_model = deepcopy(new_model)
        cloned_model = cloned_model.type(self.dtype)
        if self.gpu:
            self._model = cloned_model.cuda()
        else:
            self._model = cloned_model.cpu()

    def update_state(
            self,
            state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]],
            pseudo_state: xr.DataArray,
            analysis_time: pd.Timestamp
    ) -> xr.DataArray:
        """
        This method updates given `state` with given `observations`. This method
        stacks the observations and covariance together. The state values and
        observations are localized to given `analysis_time`, if ``smoother`` is
        set to False. In an additional step, the prepared states are
        transferred to :py:class:`torch.tensor`, on either CPU or GPU, depending
        on set `gpu` flag. :py:meth:`self.model.assimilate` is called for
        assimilation with prepared state, observations and observation
        covariance. Estimated analysis is converted into a
        :py:class:`xarray.DataArray`, which is returned.

        Parameters
        ----------
        state : :py:class:`xarray.DataArray`
            This state is localized in time and passed to
            :py:meth:`self.model.assimilate` as first argument. It is
            further updated by this assimilation algorithm and given
            ``observation``. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`.
        observations : :py:class:`xarray.Dataset` or \
        iterable(:py:class:`xarray.Dataset`)
            These observations are used to update given state. An iterable of
            many :py:class:`xarray.Dataset` can be used to assimilate different
            variables. For the observation state, these observations are
            stacked such that the observation state contains all observations.
        analysis_time : :py:class:`datetime.datetime`
            This analysis time determines at which point the state is updated.
            This argument is not used if `smoother` is set to True.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            The analysed state based on given state and observations. The
            analysis has same coordinates as given ``state``. If filtering mode
            is on, then the time axis has only one element.
        """
        logger.info('####### Neural network assimilation #######')
        logger.info('Preparing observations')
        obs_state, _ = self._prepare_obs(observations)
        logger.info('Transferring data to torch')
        prepared_torch = self._states_to_torch(state.values, obs_state,
                                               pseudo_state.values)
        logger.info('Assimilating observations with model')
        torch_analysis = self.model.assimilate(*prepared_torch)
        logger.info('Gathering analysis')
        if self.gpu:
            torch_analysis = torch_analysis.cpu()
        analysis = state.copy(deep=True, data=torch_analysis.numpy())
        logger.info('Finished with analysis creation')
        return analysis



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 2/13/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Iterable

# External modules
import numpy as np
import scipy.spatial.distance
import xarray as xr

# Internal modules
from . import common


logger = logging.getLogger(__name__)


__all__ = ['preprocess_cosmo', 'postprocess_cosmo']

_cosmo_vcoords = ['height_2m', 'height_10m', 'height_toa', 'soil1',
                  'level1', 'level', 'no_vgrid']


def preprocess_cosmo(
        cosmo_ds: xr.Dataset,
        assim_vars: Iterable[str]
) -> xr.DataArray:
    """
    This function can be used to pre-process COSMO data. There are different
    pre-processing steps included:

    1. Select variables to assimilate with `assim_vars`
    2. Set missing vertical levels of COSMO based on `vcoord`
    3. Interpolate all vertical levels of COSMO to a merge vertical grid
    4. Stack horizontal and vertical grid to one single dimension
    5. Transpose dimensions such that the resulting state array is valid

    Parameters
    ----------
    cosmo_ds : :py:class:`xarray.Dataset`
        From this COSMO dataset selected variables are extracted and this
        dataset is converted into a :py:class:`xarray.DataArray`. This dataset
        needs `vcoord` as data variable to determine the vertical coordinates of
        COSMO.
    assim_vars : iterable(str)
        These variables are included in the resulting and prepared array. If a
        variable cannot be found within the data, a warning will be raised.

    Returns
    -------
    prepared_data : :py:class:`xarray.DataArray`
        This array is prepared such that it could be assimilate by any
        assimilation algorithm within this package. The vertical grid is unified
        and stack together with the horizontal grid. The dataset with the
        selected variables is converted into this array, where 'var_name'
        indicates the variable axis.
    """
    avail_vars = [var for var in assim_vars if var in cosmo_ds.data_vars]
    not_avail_vars = list(set(assim_vars) - set(avail_vars))
    if not_avail_vars:
        logger.warning('Following variables are not found! {0:s}'.format(
            ', '.join(not_avail_vars)
        ))
    assim_ds = cosmo_ds[avail_vars]
    vgrid_ds = _prepare_vgrid(assim_ds, cosmo_ds['vcoord'])
    added_ds = common.add_no_vgrid(vgrid_ds, _cosmo_vcoords, 0)
    interp_ds = _interp_vgrid(added_ds)
    prepared_ds = _replace_coords(interp_ds)
    prepared_data = common.ds_to_array(prepared_ds, ('rlat', 'rlon', 'vgrid'))
    return prepared_data


def postprocess_cosmo(
        analysis_data: xr.DataArray,
        cosmo_ds: xr.Dataset
) -> xr.Dataset:
    """
    This function can be used to post-process COSMO data and incorporate
    included variables into given COSMO dataset. There are different steps
    included:

    1. Unstack the grid form analysis data and convert the data into a dataset
    2. Iterate through included variables and reindex these variables as they
        are in given COSMO dataset
    3. Replace the variables in given COSMO dataset with the analysis variables

    Parameters
    ----------
    analysis_data : :py:class:`xarray.DataArray`
        This array represents a valid analysis state, which can be a result of
        any given assimilation algorithm included in this package. In this
        analysis data, the horizontal and vertical grid are stacked together and
        the variables are a dimension within the array.
    cosmo_ds : :py:class:`xarray.Dataset`
        This COSMO dataset is used as source dataset to convert given analysis
        array into a valid COSMO dataset. The resulting dataset is a copy of
        this dataset, where the assimilated variables are replaced.

    Returns
    -------
    analysis_ds : :py:class:`xarray.Dataset`
        This analysis dataset is a copy of given COSMO dataset with replaced
        variables from given analysis array.
    """
    analysis_ds = common.generic_postprocess(
        analysis_data, cosmo_ds,
        vcoords=_cosmo_vcoords
    )
    logger.info('Finished post-processing of COSMO')
    return analysis_ds


def _prepare_vgrid(ds: xr.Dataset, vcoord: xr.DataArray) -> xr.Dataset:
    ds = ds.copy()
    dims_non_vert = [d for d in vcoord.dims if d not in _cosmo_vcoords]
    vcoord_vals = vcoord.mean(dim=dims_non_vert).values
    if 'soil1' in ds.coords:
        ds['soil1'] = ds['soil1'].copy(data=ds['soil1']*(-1))
        vgrid_coords = np.concatenate([vcoord_vals, ds['soil1'].values])
    else:
        vgrid_coords = vcoord_vals
    ds = ds.assign_coords(vgrid=vgrid_coords)
    if 'level1' in ds.dims:
        ds['level1'] = vcoord_vals
    if 'level' in ds.dims:
        ds['level'] = ((vcoord_vals+np.roll(vcoord_vals, 1))/2)[1:]
    return ds


def _interp_vgrid(ds: xr.Dataset) -> xr.Dataset:
    vgrid_neighbor_funcs = {
        'no_vgrid': _inds_nearest,
        'height_2m': _inds_nearest,
        'height_10m': _inds_nearest,
        'height_toa': _inds_nearest,
        'soil1': _inds_bottom,
        'level1': _inds_top,
        'level': _inds_top
    }
    vertical_coords = [c for c in _cosmo_vcoords if c in ds.coords]
    for c in vertical_coords:
        vgrid_inds = vgrid_neighbor_funcs[c](ds[c].values, ds['vgrid'].values)
        ds[c] = ds[c].copy(data=ds['vgrid'].values[vgrid_inds])
        ds = ds.reindex(**{c: ds['vgrid'].values}, method=None)
    return ds


def _inds_nearest(coord_val: np.ndarray, vgrid_val: np.ndarray) -> np.ndarray:
    dist_matrix = scipy.spatial.distance.cdist(
        coord_val[:, None], vgrid_val[:, None]
    )
    dist_argmin = np.argmin(dist_matrix, axis=1)
    return dist_argmin


def _inds_top(coord_val: np.ndarray, vgrid_val: np.ndarray) -> np.ndarray:
    return np.arange(len(vgrid_val))[:len(coord_val)]


def _inds_bottom(coord_val: np.ndarray, vgrid_val: np.ndarray) -> np.ndarray:
    return np.arange(len(vgrid_val))[-len(coord_val):]


def _replace_coords(ds: xr.Dataset) -> xr.Dataset:
    rename_vertical = {c: 'vgrid' for c in _cosmo_vcoords}
    ds = common.replace_grid(ds, rename_vertical)
    rename_horizontal = {'srlat': 'rlat', 'srlon': 'rlon'}
    ds = common.replace_grid(ds, rename_horizontal)
    return ds



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch
import torch.nn

# Internal modules
from .base_kernels import BaseKernel
from .utils import dot_product


logger = logging.getLogger(__name__)


__all__ = ['TanhKernel']


class TanhKernel(BaseKernel):
    """
    This tangens hyperbolicus kernel represents a multilayer perceptron
    kernel and often also called sigmoid kernel :cite:`lin_study_2003`. The
    dot product of the data is multiplied a coefficient :math:`\\alpha` and
    an additional intercept constant :math:`c` is added, before the product
    is transformed by a tanh function, which is a scaled sigmoid,


    .. math::

       K(x_i, x_j) = \\tanh(\\alpha {(x_i)}^{T} x_j + c)


    Parameters
    ----------
    coeff : torch.Tensor, optional
        This coefficient is used to scale the dot product output and is set
        to 1, resulting in a not scaled dot product.
    const : torch.Tensor, optional
        This constant intercept value is used to shift the output of the dot
        product, before the tanh function is applied. The default value of 0
        deactivates shifting.

    """
    def __init__(self, coeff: torch.Tensor = torch.tensor(1.),
                 const: torch.Tensor = torch.tensor(0.)):
        super().__init__()
        self.coeff = coeff
        self.const = const

    def __str__(self) -> str:
        return 'TanhKernel({0}, {1})'.format(
            str(self.coeff), str(self.const)
        )

    def __repr__(self) -> str:
        return 'Tanh({0}, {1})'.format(repr(self.coeff), repr(self.const))

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        xy = dot_product(x, y)
        logit = self.coeff * xy + self.const
        kernel_mat = torch.tanh(logit)
        return kernel_mat



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/7/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import warnings
from typing import Tuple, Union, Callable, Any

# External modules
import numpy as np

# Internal modules
from .localization import BaseLocalization


logger = logging.getLogger(__name__)


class GaspariCohn(BaseLocalization):
    """
    This localization can  be used to constrain observations. It is based on
    Gaspari-Cohn correlation function :cite:`gaspari_construction_1999`. This
    correlation function corresponds to a form factor of :math:`\\frac{1}{2}`,
    in :cite:`gaspari_construction_1999` :math:`C_0(z, \\frac{1}{2}, c)`.

    Parameters
    ----------
    length_scale : float or tuple(float)
        This length scale is :math:`c` in :cite:`gaspari_construction_1999`.
        This length scale determines
        the truncation to zero and the decay of the covariance function. The
        Gaspari-Cohn function is truncated to zero by 2 * length_scale.
    dist_func : func
        This distance function is used to determine the distance between states.
        This functions takes two different grid lists and estimates a distance
        between these two grids.
    """
    def __init__(
            self,
            length_scale: Union[float, Tuple[float]],
            dist_func: Callable,
            epsilon: float = 1E-5
    ):
        self.radius = np.atleast_1d(length_scale)
        self.dist_func = dist_func
        self.epsilon = epsilon
        self._thres = [2, 1]

    def __str__(self) -> str:
        return 'GaspariCohn(l={0})'.format(str(self.radius))

    def __repr__(self) -> str:
        return 'GaspariCohn'

    @staticmethod
    def _f1(dist: np.ndarray) -> np.ndarray:
        f1 = - 0.25 * dist ** 5
        f1 += 0.5 * dist ** 4
        f1 += 0.625 * dist ** 3
        f1 -= 5 / 3 * dist ** 2
        f1 += 1
        return f1

    @staticmethod
    def _f2(dist: np.ndarray) -> np.ndarray:
        f2 = 1 / 12 * dist ** 5
        f2 -= 0.5 * dist ** 4
        f2 += 0.625 * dist ** 3
        f2 += 5 / 3 * dist ** 2
        f2 -= 5 * dist
        f2 += 4
        f2 -= 2 / 3 / dist
        return f2

    def localize_obs(
            self,
            grid_ind: Any,
            obs_grid: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method creates weights for observations based on given grid index
        and observation grid.

        Parameters
        ----------
        grid_ind : any
            This parameter indicates the current grid index.
        obs_grid : :py:class:`np.ndarray`
            This observation grid is used to estimate a spatial distance to
            given grid index.

        Returns
        -------
        use_obs : :py:class:`np.ndarray`, dtype=bool
            This is a boolean array indicating if the `i`-th observation should
            be used. This array is calculated based on estimated observation
            weights.
        obs_weights : :py:class:`np.ndarray`, dtype=float
            The estimated observation weights. These weights can be used to
            weight observations.
        """
        weights = np.ones((obs_grid.shape[0]), dtype=float)
        dist = np.atleast_2d(self.dist_func(grid_ind, obs_grid))
        for i, d in enumerate(dist):
            dist_radius = d / self.radius[i]
            conds = [dist_radius < thres for thres in self._thres]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                tmp_weights = np.zeros((obs_grid.shape[0]), dtype=float)
                tmp_weights[conds[0]] = self._f2(dist_radius[conds[0]])
                tmp_weights[conds[1]] = self._f1(dist_radius[conds[1]])
            weights *= tmp_weights
        use_obs = weights > self.epsilon
        return use_obs, weights


class GaspariCohnInf(BaseLocalization):
    """
    This localization can  be used to constrain observations. It is based on
    Gaspari-Cohn correlation function :cite:`gaspari_construction_1999`. This
    correlation function corresponds to a form factor of infinity, in
    :cite:`gaspari_construction_1999` :math:`C_0(z, \\infty, c)`.

    Parameters
    ----------
    length_scale : float or tuple(float)
        This length scale is :math:`c` in :cite:`gaspari_construction_1999`.
        This length scale determines
        the truncation to zero and the decay of the covariance function. The
        Gaspari-Cohn function is truncated to zero by 2 * length_scale.
    dist_func : func
        This distance function is used to determine the distance between states.
        This functions takes two different grid lists and estimates a distance
        between these two grids.
    """
    def __init__(
            self,
            length_scale: Union[float, Tuple[float]],
            dist_func: Callable,
            epsilon: float = 1E-5
    ):
        self.radius = length_scale
        self.dist_func = dist_func
        self.epsilon = epsilon
        self._thres = [2, 1.5, 1, 0.5]

    def __str__(self) -> str:
        return 'GaspariCohnInf(l={0})'.format(str(self.radius))

    def __repr__(self) -> str:
        return 'GaspariCohnInf'

    @staticmethod
    def _f1(dist: np.ndarray) -> np.ndarray:
        f1 = -28 * dist ** 5 / 33
        f1 += 8 * dist ** 4 / 11
        f1 += 20 * dist ** 3 / 11
        f1 -= 80 * dist ** 2 / 33
        f1 += 1
        return f1

    @staticmethod
    def _f2(dist: np.ndarray) -> np.ndarray:
        f2 = 20 * dist ** 5 / 33
        f2 -= 16 * dist ** 4 / 11
        f2 += 100 * dist ** 2 / 33
        f2 -= 45 * dist / 11
        f2 += 51 / 22
        f2 -= 7 / (44 * dist)
        return f2

    @staticmethod
    def _f3(dist: np.ndarray) -> np.ndarray:
        f3 = -4 * dist ** 5 / 11
        f3 += 16 * dist ** 4 / 11
        f3 -= 10 * dist ** 3 / 11
        f3 -= 100 * dist ** 2 / 33
        f3 += 5 * dist
        f3 -= 61 / 22
        f3 += 115 / (132 * dist)
        return f3

    @staticmethod
    def _f4(dist: np.ndarray) -> np.ndarray:
        f4 = 4 * dist ** 5 / 33
        f4 -= 8 * dist ** 4 / 11
        f4 += 10 * dist ** 3 / 11
        f4 += 80 * dist ** 2 / 33
        f4 -= 80 * dist / 11
        f4 += 64 / 11
        f4 -= 32 / (33 * dist)
        return f4

    def localize_obs(
            self,
            grid_ind: Any,
            obs_grid: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method creates weights for observations based on given grid index
        and observation grid.

        Parameters
        ----------
        grid_ind : any
            This parameter indicates the current grid index.
        obs_grid : :py:class:`np.ndarray`
            This observation grid is used to estimate a spatial distance to
            given grid index.

        Returns
        -------
        use_obs : :py:class:`np.ndarray`, dtype=bool
            This is a boolean array indicating if the `i`-th observation should
            be used. This array is calculated based on estimated observation
            weights.
        obs_weights : :py:class:`np.ndarray`, dtype=float
            The estimated observation weights. These weights can be used to
            weight observations.
        """
        weights = np.zeros((obs_grid.shape[-1]), dtype=float)
        dist = self.dist_func(grid_ind, obs_grid)
        dist_radius = dist / self.radius
        conds = [dist_radius < thres for thres in self._thres]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            weights[conds[0]] = self._f4(dist_radius[conds[0]])
            weights[conds[1]] = self._f3(dist_radius[conds[1]])
            weights[conds[2]] = self._f2(dist_radius[conds[2]])
        weights[conds[3]] = self._f1(dist_radius[conds[3]])
        use_obs = weights > self.epsilon
        return use_obs, weights



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 10/4/18
#
# Created for neural_assim
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc
from typing import Union, Any, Tuple, Dict

# External modules
import xarray as xr
import numpy as np
import torch

# Internal modules


logger = logging.getLogger(__name__)


class BaseOperator(object):
    def __init__(
            self,
            len_grid: int = 40,
            random_state: Union[None, np.random.RandomState] = None
    ):
        """
        This is a BaseClass for observation operators. These observation
        operators are used to map a model state to observations.

        Parameters
        ----------
        len_grid : int, optional
            Number of grid points in Lorenz '96 model. Default is 40.
        random_state : :py:class:`numpy.random.RandomState` or None, optional
            This random state can be used for random numbers. Default is None.
        """
        self.len_grid = len_grid
        self.random_state = random_state

    def __call__(
            self,
            obs_ds: xr.Dataset,
            input_vals: xr.DataArray,
            *args: Tuple[Any],
            **kwargs: Dict[str, Any]
    ) -> xr.DataArray:
        pseudo_obs = self.obs_op(input_vals, *args, **kwargs)
        pseudo_obs = pseudo_obs.rename(grid='obs_grid_1')
        pseudo_obs['time'] = obs_ds.time.values
        pseudo_obs['obs_grid_1'] = obs_ds.obs_grid_1.values
        return pseudo_obs

    @abc.abstractmethod
    def obs_op(
            self,
            in_array: xr.DataArray,
            *args: Tuple[Any],
            **kwargs: Dict[str, Any]
    ) -> xr.DataArray:
        pass

    @abc.abstractmethod
    def torch_operator(self) -> torch.Tensor:
        pass



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 10.08.20
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Tuple, Any

# External modules
import torch
import dask.array as da
import numpy as np

# Internal modules


logger = logging.getLogger(__name__)


def evd(
        tensor: torch.Tensor,
        reg_value: torch.Tensor = torch.tensor(0),
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Performs eigendecomposition of a symmetric hermitian tensor. The eigenvalues
    of the nearest positive semidefinit matrix are returned to ensure
    differentiability.

    Parameters
    ----------
    tensor : :py:class:`torch.Tensor` (..., nx, nx)
        This tensor is eigen decomposed. The values of the nearest positive
        semidefinit matrix to this tensor are returned.
    reg_value : :py:class:`torch.Tensor`, optional
        This regularization value is added to the eigenvalues and represents
        a regularization of the matrix diagonal. The regularization is
        deactivated with the default value of 0.

    Returns
    -------
    evals : :py:class:`torch.Tensor` (..., nx)
        The eigenvalues of the nearest positive semidefinit matrix to the
        given tensor. The regularization value is already added to these
        eigenvalues.
    evects : :py:class:`torch.Tensor` (..., nx, nx)
        The estimated eigenvectors based on given tensor.
    evals_inv : :py:class:`torch.Tensor` (..., nx)
        The inverted eigenvalues, which can be used to invert the given tensor.
    """
    evals, evects = torch.symeig(tensor, eigenvectors=True, upper=False)
    evals = evals.clamp(min=0)
    evals = evals + reg_value
    evals_inv = 1 / evals
    return evals, evects, evals_inv


def rev_evd(
        evals: torch.Tensor,
        evects: torch.Tensor,
) -> torch.Tensor:
    """
    Composes a tensor :math:`A` based on given eigenvalues :math:`\\lambda` and
    eigenvectors :math:`u`,


    .. math::

       A = u \\lambda (u)^T


    Parameters
    ----------
    evals : :py:class:`torch.Tensor` (..., nx)
        These eigenvalues are used to recompose the matrix.
    evects : :py:class:`torch.Tensor` (..., nx, nx)
        These eigenvectors are used to recompose the matrix.

    Returns
    -------
    rev_mat : :py:class:`torch.Tensor` (..., nx, nx)
        The recomposed matrix based on given eigenvalues and eigenvectors.
    """
    diag_flat_evals = torch.diag_embed(evals)
    rev_mat = torch.einsum('...ij,...jk->...ik', evects, diag_flat_evals)
    rev_mat = torch.einsum('...ij,...kj->...ik', rev_mat, evects)
    return rev_mat


def grid_to_array(
        index: Any
) -> np.ndarray:
    """
    Transform a given index into a :py:class:`np.ndarray`, which can be then
    used by localization within the assimilation.

    Parameters
    ----------
    index : any
        This index is transformed into an array.

    Returns
    -------
    index_array : :py:class:`np.ndarray` (n_points, n_grid)
        This is the transformmed array, which is then used within the
        assimilation.
    """
    if isinstance(index, np.ndarray):
        raw_index_array = index
    elif isinstance(index, da.Array):
        raw_index_array = index.compute()
    else:
        raw_index_array = np.atleast_1d(index.values)
    if isinstance(raw_index_array[0], tuple):
        shape = (-1, len(raw_index_array[0]))
    elif raw_index_array.ndim > 1:
        shape = raw_index_array.shape
    else:
        shape = (-1, 1)
    dtype = ','.join(['float']*shape[-1])
    index_array = np.array(index, dtype=dtype).view(float).reshape(*shape)
    return index_array



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 14.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import types
from inspect import signature

# External modules
import xarray as xr
from xarray import register_dataset_accessor

# Internal modules


logger = logging.getLogger(__name__)


class ObservationError(Exception):  # pragma: no cover
    """
    This error is an error if given observation is not valid or if there is
    something strange with the observations.
    """
    pass


@register_dataset_accessor('obs')
class Observation(object):
    """
    This class represents an observation subset. Within an observation subset,
    the observations can be spatial correlated. Different observation subsets
    are per definition uncorrelated. This class is registered under
    :py:attr:`xarray.Dataset.obs`, where the here defined attributes and methods
    can be accessed. The assimilation algorithms will use the abstract method
    :py:meth:`xarray.Dataset.obs.operator` to convert the given model field into
    an observation equivalent. Thus, the `operator` method needs to be
    overwritten for every observation subset independently.

    Arguments
    ---------
    xr_ds : :py:class:`~xarray.Dataset`
        This :py:class:`~xarray.Dataset` is used for the observation operator.
        The dataset needs two variables:

            observations
                (time, obs_grid_1), the actual observation values

            covariance
                (obs_grid_1) or (obs_grid_1, obs_grid_2), the covariance between
                different observations. If this is a vector, then it is assumed
                that the observations are uncorrelated and only the variances
                are witihn this array.

        ``obs_grid_1`` and ``obs_grid_2`` are the same, but due to internals of
        xarray they are saved under different coordinates. It is possible to
        define different observation times within the `time` coordinate of the
        given :py:class:`~xarray.Dataset`.

    Warnings
    --------
    **To use this observation subset, you need to overwrite the observation
    operator**
    """
    def __init__(self, xr_ds: xr.Dataset):
        self.ds = xr_ds

    def __str__(self):
        return 'Obs dataset ({0})'.format(str(self.ds))

    def __repr__(self):
        return 'Observation'

    @property
    def correlated(self) -> bool:
        """
        Checks if the observations are correlated based on the dimensions of
        given dataset.

        Returns
        -------
        correlated : bool
            If the observations are correlated
        """
        correlated = 'obs_grid_2' in self.ds['covariance'].dims
        return correlated

    @property
    def _valid_dims(self) -> bool:
        """
        Checks if ``time``, ``obs_grid_1`` are available within the
        :py:class:`~xarray.Dataset` and if ``time``, ``obs_grid_1`` and
        ``obs_grid_2`` are the only dimensions.

        Returns
        -------
        valid_dims : bool
            If the dimensions are available.
        """
        necessary_dims = [
            'time', 'obs_grid_1'
        ]
        keys_avail = all(
            True if d in tuple(self.ds.dims.keys()) else False
            for d in necessary_dims
        )

        all_dims = necessary_dims + ['obs_grid_2']
        no_auxiliary = all(
            True if d in all_dims else False for d in tuple(self.ds.dims.keys())
        )

        valid_dims = keys_avail and no_auxiliary
        return valid_dims

    @property
    def _valid_obs(self) -> bool:
        """
        Checks if dimensions of the ``observation``
        :py:class:`~xarray.DataArray` within the set dataset are valid.

        Returns
        -------
        valid_obs : bool
            If the ``observation`` :py:class:`~xarray.DataArray` is valid.
        """
        valid_dims = ('time', 'obs_grid_1')
        valid_obs = valid_dims == self.ds['observations'].dims[-2:]
        return valid_obs

    @property
    def _valid_cov_uncorr(self) -> bool:
        """
        Checks if shape and dimensions of the uncorrelated ``covariance``
        :py:class:`~xarray.DataArray` within the set dataset are valid.

        Returns
        -------
        valid_cov : bool
            If the ``covariance`` :py:class:`~xarray.DataArray` is valid.
        """
        dim_order = ('obs_grid_1',)
        checked_dims = dim_order == self.ds['covariance'].dims

        obs_grid_len = self.ds['observations'].shape[-1]
        valid_shape = (obs_grid_len,)
        checked_shape = valid_shape == self.ds['covariance'].shape
        valid_cov = checked_dims and checked_shape
        return valid_cov

    @property
    def _valid_cov_corr(self) -> bool:
        """
        Checks if shape and dimensions of the correlated ``covariance``
        :py:class:`~xarray.DataArray` within the set dataset are valid.

        Returns
        -------
        valid_cov : bool
            If the ``covariance`` :py:class:`~xarray.DataArray` is valid.
        """
        dim_order = ('obs_grid_1', 'obs_grid_2')
        checked_dims = dim_order == self.ds['covariance'].dims[-2:]

        obs_grid_len = self.ds['observations'].shape[-1]
        valid_shape = (obs_grid_len, obs_grid_len)
        checked_shape = valid_shape == self.ds['covariance'].shape

        try:
            checked_coord_values = self.ds['obs_grid_1'].to_index().equals(
                self.ds['obs_grid_2'].to_index()
            )
        except KeyError:
            checked_coord_values = False

        valid_cov = checked_dims and checked_shape and checked_coord_values
        return valid_cov

    @property
    def _valid_arrays(self) -> bool:
        """
        Checks if ``observations`` and ``covariance``
        :py:class:`~xarray.DataArray`s within the set dataset are valid.

        Returns
        -------
        valid_arrays : bool
            If the two :py:class:`~xarray.DataArray`s are valid.
        """
        try:
            if self.correlated:
                valid_array = self._valid_obs and self._valid_cov_corr
            else:
                valid_array = self._valid_obs and self._valid_cov_uncorr
        except KeyError:
            valid_array = False
        return valid_array

    @property
    def valid(self) -> bool:
        """
        Checks if set :py:class:`~xarray.Dataset` is valid.

        Returns
        -------
        valid_ds : bool
            If set :py:class:`~xarray.Dataset` is valid.
        """
        valid_ds = False
        if self._valid_dims and self._valid_arrays:
            valid_ds = True
        return valid_ds

    @staticmethod
    def operator(obs_ds: xr.Dataset, state: xr.DataArray) -> xr.DataArray:
        raise NotImplementedError('No observation operator is set!')
